/*   Always uses explicit negs for demo. */
/*   Updated to include MAO as input.  Means need to make a new netfile.  */

/*
 *   Non-punctate case.  Ripped off from ../Regions/play.c
 *
 *   This is actually not all that different.  The easiest way to enforce
 *     weight-tying on the massive scale I want it, while not producing a
 *     memory-hog, is to have the architecture consist of only a single
 *     RF, and have the rest of the stuff built into the simulator, so it
 *     switches from one (virtual) RF to another, and does all the book
 *     keeping necessary for backprop.
 *
 *   Two new types of nodes:  ScanningNode and MaxScanningNode.  SN scans
 *     an RF over all TR boundary positions, and returns the average value;
 *     MSN returns the max.  
 *
 *   For now, assuming learning one concept at a time.  See GetScenes for
 *     loading the goal vector when reading in the scenes.
 */

/*   Uses a new improved learning rule for scanning nodes.....  */

#include <stdio.h>
#include <dirent.h>
#include "X11/Intrinsic.h"
#include "X11/StringDefs.h" 
#include "X11/Shell.h" 
#include "X11/Xaw/Command.h"
#include "X11/Xaw/Label.h"
#include "X11/Xaw/Box.h"
#include "X11/Xaw/Cardinals.h"
#include "X11/Xaw/Dialog.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <sys/types.h>
#include <math.h>

#include "Bitmaps/zero.b"
#include "Bitmaps/one.b"
#include "Bitmaps/two.b"
#include "Bitmaps/three.b"
#include "Bitmaps/four.b"
#include "Bitmaps/five.b"
#include "Bitmaps/six.b"
#include "Bitmaps/seven.b"
#include "Bitmaps/eight.b"
#include "Bitmaps/nine.b"
#include "Bitmaps/ten.b"

#include "Bitmaps/LL.b"
#include "Bitmaps/UL.b"
#include "Bitmaps/UR.b"
#include "Bitmaps/LR.b"

#include "Bitmaps/box.b"
#include "Bitmaps/mark.b"
#include "Bitmaps/front.b"

#define PI    3.1415927
#define FALSE 0
#define TRUE 1

#define DIM 50             /* Object image is DIMxDIM. */
#define SDIM 150           /* Scene is SDIMxSDIM */

#define N 100              /* Max number of units allowed in net */
#define NLMS 15            /* Max number of landmarks used in training */
#define MAXTHRU 10         /* Max number units going through points */
#define MAXTIED 20         /* Max number of classes of tied weights */
#define MAXFIXED 20        /* Max number of fixed weights allowed */
#define MAXSCENES 90       /* Max number of scenes to train on. */

typedef struct
{
  float Sine;
  float Cosine;
} Angle;

/*  X-related global variables */

Display              *disptr;
Window               win; 
XEvent               event;
GC                   gc;
XFontStruct          *font;
XGCValues            values;
XWindowAttributes    attr;
XSetWindowAttributes set_attr;
XImage               img, *AnImage;
XColor               Grey[256];          /* Grey scale */
Pixmap               pmap;

static char *Bbits[] =
{
  zero_bits, one_bits, two_bits, three_bits, four_bits, five_bits,
  six_bits, seven_bits, eight_bits, nine_bits, ten_bits
};

static int Bwidth[] =
{
  zero_width, one_width, two_width, three_width, four_width, five_width,
  six_width, seven_width, eight_width, nine_width, ten_width
};

static int Bheight[] =
{
  zero_height, one_height, two_height, three_height, four_height, five_height,
  six_height, seven_height, eight_height, nine_height, ten_height
};

/*  Other global variables */

int   Epoch;               /* Current epoch number */
float WeightRange;         /* Random-init weights in range [-WR,+WR] */
float SigmoidPrimeOffset;  /* Add to sigmoid-prime to kill flat spots */
int   HyperErr;            /* 1 => use atanh error function */
int   SplitEpsilon;        /* 1 => divide epsilon by fan-in before use */
float Epsilon;             /* For grad descent if last step was (almost) 0 */
float Momentum;            /* Normal old momentum term */
float ModeSwitchThreshold; /* Inside thresh, do grad descent; outside, jump. */
float MaxFactor;           /* Don't jump more than this times last step */
float Decay;               /* Weight decay */
int   SinglePass;          /* 1 => Pause after forward/backward cycle */
int   SingleEpoch;         /* 1 => Pause after each training epoch */
int   Step;                /* Turned to 1 after each pause, briefly */
int   Restart;             /* 1 => restart when max epochs reached */
int   KeepScore;           /* 1 => accumulate error score for each epoch */
float TotalError;          /* Total output error for one epoch */
float ScoreThreshold;      /* This close to desired value => bit is correct */
int   TotalErrorBits;      /* Total # bits in epoch that were wrong */
int   PosErrorBits;        /* Total # wrong positive instances in epoch */
int   DidGradient;         /* Total # patterns that did gradient descent */

int   Nunits;              /* Total number of units in net */
int   Ninputs;             /* Number of input units */
int   FirstHidden;         /* Index of 1st hidden unit */
int   Nhidden;             /* Number of hidden units */
int   FirstOutput;         /* Index of 1st output unit */
int   Noutputs;            /* Number of output units */

float Outputs[N];          /* Final output value for each unit */
float ErrorSums[N];        /* Total error activation for each unit */
float Errors[N];           /* Final error value for each unit */
int   Nconnections[N];     /* # of INCOMING connections per unit */
int   Connections[N][N];   /* C[i][j] lists jth unit projecting to unit i */
float Weights[N][N];       /* W[i][j] holds weight of C[i][j] */
float DeltaWeights[N][N];  /* Change between previous weight and current one */
float Slopes[N][N];        /* Accumulated slope value for each position */
float PrevSlopes[N][N];    /* Similarly, for the last position visited */

int   NTrainingPatterns;   
int   NPositiveTrainingPatterns;
int   NTestPatterns;       
float TrainingInputs[1000][N];
float TrainingOutputs[1000][N];
float TestInputs[N][N];
float TestOutputs[N][N];

float Tinputs[N];          /* Input vector to be tested. */
float AvgOutputs[N][N];       /* Outputs averaged over position */
float ScanningSum[N][N];      /* sum of fprime * o_i^k, see thesis. */
float MaxResponse[N];
float OutputsforMaxResponse[N][N];    /* Output values of *input* nodes that
					 yielded max overall response. */

/* Angle node stuff... RBF nodes */

int AngleNode[N];
float DistSquared[N];      /* Distance squared from canon dir for RBF */

float SigmaDeriv;          /* Value for sigma to use when computing
			      derivative for sin,cos */

/* Averaging, scanning, and max-scanning node stuff */

int AveragingNode[N];
int ScanningNode[N];
int MaxScanningNode[N];

/* Weight tying */

int NumTied;
int TiedClassSize[MAXTIED];  /* Size of tie-class */
int Tied[MAXTIED][N][2];     /* Tied[class][ix][0/1] holds that either the
				0th or 1st element of the pair defining the
				ix'th member of the class.  A pair is defined:
				0:  the node the link comes into.
				1:  the index of the link coming in.  */

/* Weight fixing */

int NumFixed;
int Fixed[N][N];

/* Weights from inputs */

int NumInputWeights;
int InputWeights[MAXTIED][3];   /* 0: index of hidden node, 
				   1: index of incoming link, 
				   2: input node to get weight value from. */

/* Object stuff */

int TRx[MAXSCENES],TRy[MAXSCENES];
int LMx[MAXSCENES],LMy[MAXSCENES];
char CurObjfname[80];
XImage CurObjimage;
XImage TRimage[MAXSCENES];
XImage LMimage[MAXSCENES];
int    TRBoundaryPoints[MAXSCENES][DIM*DIM][2]; /* pt is a TR boundary pt? */
int    NumTRBoundaryPoints[MAXSCENES];          /* how many are there? */

int   Interior[NLMS][DIM][DIM];   /* LM interior */
float HalfLength[NLMS];   /* Half the length of the major axis */
float HalfWidth[NLMS];    /* Half the width (minor axis) */
Angle MajorAxisOrientation[MAXSCENES][2];  /* 2 since dir ambig, +/- theta */
Angle MinorAxisOrientation[MAXSCENES][2];  /* 2 since dir ambig, +/- theta */
Angle ProxOrientation[MAXSCENES];     /* LM to TR ray */
Angle FrontOrientation[NLMS][2];  /* 2 since direction ambig, +/- theta */
int   Spread[SDIM][SDIM];   /* Spread map */
int   ClosestX, ClosestY; /* LM point closest to TR */
int   PatClosestX[1000], PatClosestY[1000];   /* closest for a seen pattern */
int   SceneIndex;         /* Which scene are we looking at? */

int ExplicitNegatives;    /* Use explicit negatives? */
int DoAntonyms;           /* Use antonyms? */
int NumLMs;               /* Number of LMs used. */
int DisplayMe;            /* Display this output lexeme: 0 [default] or 1 */
int StartingEpox;         /* Number of epochs to start training with */
int ZeroStop;             /* Stop when TotalErrorBits == 0. */
int PosZeroStop;          /* Stop when PosErrorBits == 0. */
int FixCoM;               /* Freeze the CoM weights at horiz, vert? */
int Which[1000];          /* Which LM the training pattern goes with (0/1). */
int LM_index;             /* LM_index = Which[i] */
int Freeze;               /* Flag:  Freeze some weights? */
int FreezeStart;          /* Beginning of range of weights to be frozen */
int FreezeStop;           /* End of range of weights to be frozen */
float ErrorCut;           /* Reduce error on negative instances */

float Beta[N][N];         /* Betas for error reduction */
float Atten;              /* Amount to atten dg/dx for sin,cos consistency */

int Antonyms[N];          /* Antonyms for each output unit */

int Window_width;
int Window_height;

int X_off, Y_off;

char archfname[80];

/*  Xt stuff, see apology below ("kind of gross").  */

Widget toplevel;
static XrmOptionDescRec options[] = 
{
    {"-hspace",	"*Box.hSpace",	XrmoptionSepArg,	NULL},
    {"-vspace",	"*Box.vSpace",	XrmoptionSepArg,	NULL},
};

main(argc,argv)
     int argc;
     char *argv[];
{
  int ing;                  /* Needed for command line parsing */
  extern char *optarg;      /* Needed for command line parsing */

  unsigned int seed;
  char state[256];
  /*  long random();*/

  float f;
  int old_weights;
  int i, j, k, epox, response;
  float divisor;
  float RandomWeight(), AspectRatio();
  char fname[80];
  char conceptname[80];
  char testname[80];
  char pos_res_fname[80];         /* File for results from +ive instances */
  char neg_res_fname[80];         /* File for results from -ive instances */
  char cmdname[80];
  char inc;
  FILE  *pos_res_file, *neg_res_file, *fopen();

  int bitmap_size;
  int want_to_train;
  int want_to_test_individuals;
  int test_set;

  int status;
  unsigned int *width_ret, *height_ret;
  int *x_hot_ret, *y_hot_ret;

  float x,y, offsin,offcos;    /* , DtoR(); */
  float ftx,fty,flx,fly;
  float rhx,rhy,rfx,rfy,lhx,lhy,lfx,lfy,midheadx,midheady,midfootx,midfooty;

  unsigned char BitmapBit();

  Angle *directionptr;

  XtAppContext app_con;

  directionptr = (Angle *) malloc (sizeof(Angle));

  seed = 0;
  want_to_train = FALSE;
  want_to_test_individuals = FALSE;
  test_set = FALSE;
  archfname[0] = '*';
  fname[0] = '*';
  conceptname[0] = '*';
  old_weights = FALSE;
  NumLMs = 1;              /* Default */
  DisplayMe = 0;
  StartingEpox = 0;
  FixCoM = FALSE;
  ZeroStop = FALSE;
  PosZeroStop = FALSE;
  ErrorCut = 1.0;
  DoAntonyms = FALSE;
  ExplicitNegatives = TRUE;     /* Always explicit negs for demo. */
  Atten = 1.0;
  SigmaDeriv = 3.0;

  Freeze = FALSE;
  FreezeStart = 0;
  FreezeStop = 0;


  /* Initialize */

  InitializeGlobals();

  /* Read in options */

  while ((ing = getopt(argc,argv,"a:p:c:or:tT:im:#:d:s:fzZ[:]:e:E:A:S:")) 
	 != EOF)
    switch (ing)
      {
      case 'a': strcpy(archfname, optarg);                  break;
      case 'p': strcpy(fname, optarg);                      break;
      case 'c': strcpy(conceptname, optarg);                break;
      case 'o': old_weights = TRUE;                         break;
      case 'r': seed = (unsigned int) atoi(optarg);         break;
      case 't': want_to_train = TRUE;                       break;
      case 'T': test_set = TRUE; strcpy(testname,optarg);   break;
      case 'i': want_to_test_individuals = TRUE;            break;
      case 'm': ModeSwitchThreshold = atof(optarg);         break;
      case '#': NumLMs = atoi(optarg);                      break;
      case 'd': DisplayMe = atoi(optarg);                   break;
      case 's': StartingEpox = atoi(optarg);                break;
      case 'f': FixCoM = TRUE;                              break;
      case 'z': ZeroStop = TRUE;                            break;
      case 'Z': PosZeroStop = TRUE;                         break;
      case '[': Freeze = TRUE; FreezeStart = atoi(optarg);  break;
      case ']': Freeze = TRUE; FreezeStop = atoi(optarg);   break;
      case 'e': ErrorCut = atof(optarg);                    break;
      case 'E': Epsilon = atof(optarg);                     break;
      case 'A': Atten = atof(optarg);                       break;
      case 'S': SigmaDeriv = atof(optarg);                  break;
      }

  /*
   *  Kind of gross.  Had to have this guy in here just so I could use
   *    the XmuReadBitmapFile routine (see below).   It's not a widget
   *    based program at all.
   */

  toplevel = XtAppInitialize(&app_con, "Xbuttonbox",
			     options, XtNumber(options),
			     &argc, argv, NULL, NULL, ZERO);



  InitializeBetas();

  if ( Freeze )
    printf ("Freezing incoming weights on nodes %d to %d.\n",
	    FreezeStart, FreezeStop);

  /*  Handle options */

  if (seed == 0)
    srandom(getpid());
  else 
    {
      initstate(seed, state, 65536);
      srandom(seed);
    }

  if ((conceptname[0] == '*') || (archfname[0] == '*'))
    {
      fprintf(stderr, ">> Error:  must specify concept and archfile.\n");
      fprintf(stderr, "   train -c concept -a archfile [options]\n");
      exit(1);
    }

  
  /*  Display initialization (but don't map). */

  disptr=XOpenDisplay(NULL);
  X_off = Bwidth[0];
  Y_off = Bheight[0];
  Window_width = (Bwidth[0] * DIM) + (2 * X_off);
  Window_height = (Bheight[0] * DIM) + (2 * Y_off);
  if (!(font = XLoadQueryFont(disptr, "9x15")))
    {
      printf ("Error:  Cannot load font.\n");
      exit(1);
    }
  win=XCreateSimpleWindow(disptr,RootWindow(disptr,0), 50,100,
			  Window_width, Window_height, 5,0,
			  WhitePixel(disptr,0));
  AnImage = XCreateImage(disptr,&win,
			 1, XYBitmap, 0, NULL, DIM,DIM,
			 8, ((DIM + 7) / 8));

  /* Get structure, scenes, old weights if any.  */

  GetArchitecture(archfname);

  if ( test_set )
    {
      if ( ExplicitNegatives )
	GetScenes(testname);
      else
	ImplicitNegsGetScenes(testname);
    }
  else
    {
      if ( ExplicitNegatives )
	GetScenes(conceptname);
      else
	ImplicitNegsGetScenes(conceptname);
    }

  if (old_weights)
    GetWeights(conceptname);                 /* Overrides default archfname */

  /* Train the sucker. */

  if ( want_to_train == TRUE )
    {
      /* Don't know the closest point for this pattern yet. */

      for (i=0; i<NTrainingPatterns; i++)
	{
	  PatClosestX[i] = -1;
	  PatClosestY[i] = -1;
	}

      if ( StartingEpox != 0 )
	Train( StartingEpox );

      epox = 34;
      while ( epox != 0 )
	{
	  printf ("Enter number of epochs to train: ");
	  scanf ("%d", &epox);
	  if ( epox != 0 )
	    Train ( epox );
	}
    }

  /* Test the sucker. */

  if ( want_to_test_individuals )
    Test();
  else
    {

/*
      sprintf (pos_res_fname, "%s.res", conceptname);
      sprintf (neg_res_fname, "%s-neg.res", conceptname);
*/
      pos_res_file = fopen( "pos", "w");
      neg_res_file = fopen( "neg", "w");


      /*  Get scores for all members of training set, and write 
       *    to result files ( <conceptname>.res ).   Result files
       *    are composed of lines of the form %d %f, where the integer
       *    is the scene index, and the float is the result.  End of
       *    data is marked by a -ive scene index.
       */

      for (SceneIndex=0; SceneIndex<NTrainingPatterns; SceneIndex++)
	{
	  FillTinputs();
	  ForwardPass( Tinputs );
	  if (SceneIndex < NPositiveTrainingPatterns)    /* Pos instance */
	    fprintf(pos_res_file, "%d %f\n", 
		    SceneIndex, Outputs[FirstOutput]);
	  else                                         /* Neg instance */
	    fprintf(neg_res_file, "%d %f\n", 
		    SceneIndex-NPositiveTrainingPatterns, 
		    Outputs[FirstOutput]);
	}
      
      fprintf (pos_res_file, "-9 -9");
      fprintf (neg_res_file, "-9 -9");
      fclose(pos_res_file);
      fclose(neg_res_file);


      /*  Build command names and issue.   */

      if ( test_set )
	{
	  sprintf (cmdname, "allscene -c %s -r pos", testname);
	  system(cmdname);
	}
      else
	{
	  sprintf (cmdname, "allscene -c %s -r pos", conceptname);
	  system(cmdname);
	  sprintf (cmdname, "allscene -c %s-neg -r neg", conceptname);
	  system(cmdname);
	}
    }
  if ( want_to_train )
    {
      printf ("Want to dump weights [0=>no, 1=>yes]? ");
      scanf ("%d", &response);
      if ( response )
	DumpWeights ( conceptname );
    }
}


/*
 *  Initialize the beta array, for error reduction.  Beta[i][j] (from i to
 *    j) indicates "how seriously" we take implicit negatives from i in 
 *    computing the error for j.
 */
InitializeBetas()
{

  int i,j;

  /*  Set up so everyone div by ErrorCut except Beta[i][i].  */

  for (i=0; i<N; i++)
    for (j=0; j<N; j++)
      Beta[i][j] = 1.0 / ErrorCut;

  for (i=0; i<N; i++)
    Beta[i][i] = 1.0;
}


/*
 *  Compute direction for the input given, i.e. sin and cosine.
 *    Note that a base angle is passed in.  Sin and cos are computed
 *    relative to that angle (i.e. we really compute theta - base).
 *
 *  Recall:  sin(x-y) = sin(x) cos(y) - sin(y) cos(x)
 *           cos(x-y) = cos(x) cos(y) + sin(x) sin(y)
 */
Direction(lm_x, lm_y, tr_x, tr_y, base_sin, base_cos, directionptr)
     float lm_x, lm_y, tr_x, tr_y;
     float base_sin, base_cos;
     Angle *directionptr;
{
  int k;
  float answer;
  float xdiff, ydiff, hypot, tmp_sin, tmp_cos;

  xdiff = (tr_x - lm_x);
  ydiff = (tr_y - lm_y);
  hypot = sqrt((xdiff*xdiff) + (ydiff*ydiff));

  /*  Store theta, note minus for sin, correcting for upside-down world */

  tmp_sin = ((hypot == 0.0) ? 0.0 : ydiff / hypot);
  tmp_cos = ((hypot == 0.0) ? 0.0 : xdiff / hypot);

  /*  Subtract base from current theta */

  directionptr->Sine   = (tmp_sin*base_cos) - (base_sin*tmp_cos);
  directionptr->Cosine = (tmp_cos*base_cos) + (tmp_sin*base_sin);

  /* Roundoff */

  if ((directionptr->Sine > -0.00000001) &&
      (directionptr->Sine <  0.00000001))
    directionptr->Sine = 0.0;

  if ((directionptr->Cosine > -0.00000001) &&
      (directionptr->Cosine <  0.00000001))
    directionptr->Cosine = 0.0;
}


/*
 *  Retrieve the value of a given pixel in the interior of an LM.
 *
 *  This is a little awkward because the i,j coordinates are relative to 
 *    the TR's origin, which is (TRx[i],TRy[i]) relative to the overall
 *    origin for scene i.
 *
 */
PeekInterior(scene_ix, tr_i,tr_j)
     int scene_ix, tr_i,tr_j;
{
  int i,j;         /* Coordinates relative to LM origin */

  i = (tr_i + TRx[scene_ix]) - LMx[scene_ix];
  j = (tr_j + TRy[scene_ix]) - LMy[scene_ix];

  if ((i<0) || (i>=DIM) || (j<0) || (j>=DIM))
    return (0);
  else 
    return (GetPixel(&LMimage[scene_ix], i,j));
}

/*
 *  Retrieve a pixel from an image.  Admittedly kind of ugly...  I use a
 *    function hanging off an arbitrary image (AnImage) to access pixels 
 *    in other images.  X programming at its finest.  You don't get the 
 *    function if you don't create using XCreateImage, is the problem.
 */
GetPixel(imageptr, i,j)
     XImage *imageptr;
     int i,j;
{
  return (*(AnImage->f.get_pixel)) (imageptr, i,j);
}


/*
 *  Get and initialize a network architecture. 
 */
GetArchitecture(fname)
     char fname[];
{
  FILE  *infile, *fopen();
  char junk[10][80];
  char stringjunk[80];
  char realfname[80];
  char c;
  int  temp[10], i, j, k, connect, anodes;
  float wt;

  /* Get numbers of input, hidden, and output units. */

  infile = fopen ( fname, "r" );

  c = 'c';           /* Discard leading comments */
  while (c != '#') 
    fscanf (infile, "%c", &c);
  
  /* Get numbers of inputs, hidden units, and output units */

  fscanf (infile, "%s %d %s %d %s %d", 
	  junk[0], &temp[0], junk[1], &temp[1], junk[2], &temp[2]);
  BuildDataStructures( temp[0], temp[1], temp[2] );


  /* Connect layers. */

  fscanf (infile, "%s %d", junk[0], &connect);

  for (i=0; i<connect; i++)          /* Reading ConnectLayers lines */
    {
      fscanf (infile, "%d %d %d %d",
	      &temp[0], &temp[1], &temp[2], &temp[3]);
      ConnectLayers ( temp[0], temp[1], temp[2], temp[3] );
    }

  /*  Find which nodes are angle nodes */

  for (i=0; i<Nunits; i++)      /* Assume not angle node */
    AngleNode[i] = FALSE;

  fscanf (infile, "%s %d", junk[0], &anodes);   /* How many are there? */
  for (i=0; i<anodes; i++)  
    {
      fscanf (infile, "%d", &j);
      AngleNode[j] = TRUE;
    }

  /*  Find which nodes are averaging nodes */

  for (i=0; i<Nunits; i++)      /* Assume not angle node */
    AveragingNode[i] = FALSE;

  fscanf (infile, "%s %d", junk[0], &anodes);   /* How many are there? */
  for (i=0; i<anodes; i++)  
    {
      fscanf (infile, "%d", &j);

      AveragingNode[j] = TRUE;
    }


  /*  Find which nodes are scanning nodes. */

  for (i=0; i<Nunits; i++)      /* Assume not scanning node */
    ScanningNode[i] = FALSE;

  fscanf (infile, "%s %d", junk[0], &anodes);   /* How many are there? */
  for (i=0; i<anodes; i++)  
    {
      fscanf (infile, "%d", &j);
      ScanningNode[j] = TRUE;
    }


  /*  Find which nodes are max-scanning nodes. */

  for (i=0; i<Nunits; i++)      /* Assume not max-scanning node */
    MaxScanningNode[i] = FALSE;

  fscanf (infile, "%s %d", junk[0], &anodes);   /* How many are there? */
  for (i=0; i<anodes; i++)  
    {
      fscanf (infile, "%d", &j);
      MaxScanningNode[j] = TRUE;
    }


  /*  Find out which weights are tied.  */

  fscanf (infile, "%s %d", junk[0], &NumTied);   /* How many classes ? */
  for (i=0; i<NumTied; i++)
    {
      fscanf (infile, "%d", &TiedClassSize[i]);
      for (j=0; j<TiedClassSize[i]; j++)
	fscanf(infile, "%d %d", &Tied[i][j][0], &Tied[i][j][1]);
    }

  /*  Find out which weights are fixed.  */

  for (i=0; i<Nunits; i++)      /* Assume not fixed */
    for (j=0; j<Nunits; j++)
      Fixed[i][j] = FALSE;

  fscanf (infile, "%s %d", junk[0], &NumFixed);   /* How many? */
  for (i=0; i<NumFixed; i++)
    {
      fscanf(infile, "%d %d %f", &j, &k, &wt);
      Weights[j][k] = wt;
      Fixed[j][k] = TRUE;
    }

  /*  Set some initial weights, but allow to change.  */

  fscanf (infile, "%s %d", junk[0], &anodes);   /* How many? */
  for (i=0; i<anodes; i++)
    {
      fscanf(infile, "%d %d %f", &j, &k, &wt);
      Weights[j][k] = wt;
    }
  
  /* Tie some weights to values of input nodes. */

  fscanf (infile, "%s %d", junk[0], &NumInputWeights);   /* How many? */
  for (i=0; i<NumInputWeights; i++)
    fscanf(infile, 
	   "%d %d %d", 
	   &InputWeights[i][0], 
	   &InputWeights[i][1],
	   &InputWeights[i][2]);

  fclose(infile);
}


/*
 *  Get training points.   fname is concept name.
 *  Load the TrainingOutputs vector here, too.  Assuming learning one
 *    concept at a time for now.
 */
GetScenes(fname)
     char fname[];
{
  FILE  *infile, *fopen();
  char junk[10][80];
  char lmfname[80];
  char realfname[80];
  char c;
  int  temp[10], i, j, ntp, read_ok, ix;

  /* Find appropriate files and open, positive instances first. */

  read_ok = TRUE;    ix = 0;
  while (read_ok)
    {
      sprintf(realfname,"%s.%d",fname,ix);
      read_ok = ReadScreenFile(realfname,ix);
      if (read_ok)
	{
	  TrainingOutputs[ix][0] = 1.0;
	  ix++;
	}
      else if (ix == 0)               /* No such files */
	{
	  printf ("No such concept.\n");
	  exit(1);
	}
    }

  NTrainingPatterns = ix;
  NPositiveTrainingPatterns = ix;

  /* Now read in negative instances. */

  read_ok = TRUE;    ix = 0;
  while (read_ok)
    {
      sprintf(realfname,"%s-neg.%d",fname,ix);     /* Negative examples */
      read_ok = ReadScreenFile(realfname,NTrainingPatterns+ix);
      if (read_ok)
	{
	  TrainingOutputs[NTrainingPatterns+ix][0] = 0.0;
	  ix++;
	}
      else if (ix == 0)               /* No such files */
	{
	  printf ("No negative instances for the concept.\n");
/*	  exit(1);    */
	}
    }

  NTrainingPatterns += ix;
}


/*
 *  Get training points.   fname is concept name.
 *  Load the TrainingOutputs vector here, too.  Assuming learning one
 *    concept at a time for now.
 */
ImplicitNegsGetScenes(fname)
     char fname[];
{
  FILE  *infile, *fopen();
  char junk[10][80];
  char lmfname[80];
  char realfname[80];
  char c;
  int  temp[10], i, j, ntp, read_ok, ix;

  /* Find appropriate files and open. */

  NPositiveTrainingPatterns = 0;
  NTrainingPatterns = 0;
  for (i=0; i<Noutputs; i++)  /* i indexes the output unit for the concept */
    {
      read_ok = TRUE;    ix = 0;
      while (read_ok)
	{
	  sprintf(realfname,"%s.%d.%d",fname,i,ix);
	  read_ok = ReadScreenFile(realfname,NTrainingPatterns);
	  if (read_ok)
	    {
	      for (j=0; j<Noutputs; j++)
		TrainingOutputs[NTrainingPatterns][j] = 0.0;
	      TrainingOutputs[NTrainingPatterns][i] = 1.0;
	      ix++;
	      NTrainingPatterns++;
	    }
	  else if (ix == 0)               /* No such files */
	    {
	      printf ("No such concept.\n");
	      exit(1);
	    }
	}
    }
}


/*
 *  Read in a screen file.   Note that we read in the LM *interior*, and
 *     the TR *boundary*.
 */
int ReadScreenFile (fname,ix)
     char *fname;
     int ix;
{
  FILE *infile, *maofile, *fopen();
  char c;
  char LMfilename[80], TRfilename[80], tmpfilename[80];
  int i,j, count, ijunk, ijunk2;
  char junk[80];
  float fjunk;
  int status, success;

  infile = fopen(fname, "r");
  success = (infile != NULL);
  if (success)
    {
      fscanf (infile, "LM: %d %d %s\n", &LMx[ix], &LMy[ix], LMfilename);
      sprintf(tmpfilename, "Filled%s", LMfilename);
      status = ReadBitmapFile (tmpfilename,&LMimage[ix]);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", LMfilename);
	  exit(1);
	}

      fscanf (infile, "TR: %d %d %s\n", &TRx[ix], &TRy[ix], TRfilename);
      status = ReadBitmapFile (TRfilename,&TRimage[ix]);
      if (status != BitmapSuccess)
	{
	  printf (">> Error:  Could not read object file %s.\n", TRfilename);
	  exit(1);
	}

      /*  Read in the proximal orientation.  */

      fscanf (infile, "ProxOrient: %f %f\n",
	      &(ProxOrientation[ix].Sine),
	      &(ProxOrientation[ix].Cosine));


      /*  Find all the TR boundary points that are on. */

      NumTRBoundaryPoints[ix] = 0;
      for (i=0; i<DIM; i++)
	for (j=0; j<DIM; j++)
	  if (GetPixel(&TRimage[ix], i,j) == 1)
	    {
	      TRBoundaryPoints[ix][NumTRBoundaryPoints[ix]][0] = i;
	      TRBoundaryPoints[ix][NumTRBoundaryPoints[ix]][1] = j;
	      NumTRBoundaryPoints[ix]++;
	    }

      /*  Retrieve MAO, mAO, etc. from InfoOnObjects/LMfilename */

      sprintf(tmpfilename, "InfoOn%s", LMfilename);
      maofile = fopen(tmpfilename, "r");
      success = (maofile != NULL);
      if (success)
	{
	  c = 'a';  while (c != '#') fscanf(maofile,"%c",&c);
	  fscanf(maofile, "%s %d", junk, &ijunk);     /* area */
	  fscanf(maofile, "%s %d %d", junk, &ijunk, &ijunk2);    /* CoM */
	  fscanf(maofile, "%s %f", junk, &fjunk);     /* eccentricity */
	  fscanf(maofile, "%s %f %f", 
		 junk, 
		 &MajorAxisOrientation[ix][0].Sine,
		 &MajorAxisOrientation[ix][0].Cosine);
	  fscanf(maofile, "%s %f %f", 
		 junk, 
		 &MinorAxisOrientation[ix][0].Sine,
		 &MinorAxisOrientation[ix][0].Cosine);

	  /*
	   * NOTE!!!  Temporarily assuming that LMs do not have
	   *    any directionality associated with their surfaces.
	   */

	  MajorAxisOrientation[ix][1].Sine = 
	    MajorAxisOrientation[ix][0].Sine;
	  MajorAxisOrientation[ix][1].Cosine = 
	    MajorAxisOrientation[ix][0].Cosine;
	  MinorAxisOrientation[ix][1].Sine = 
	    MinorAxisOrientation[ix][0].Sine;
	  MinorAxisOrientation[ix][1].Cosine = 
	    MinorAxisOrientation[ix][0].Cosine;
	  fclose(maofile);
	}

      fclose(infile);
    }

  return ( success );
}


/*
 *  Read in a bitmap representation of an object.
 */
int ReadBitmapFile (filename,image)
     char *filename;
     XImage *image;
{
    unsigned int width, height;
    int x_hot, y_hot;
    unsigned char *data;
    int status;

    status = XmuReadBitmapDataFromFile (filename, &width, &height, &data,
					&x_hot, &y_hot);
    if (status != BitmapSuccess) return status;

    image->width = width;
    image->height = height;
    image->xoffset = 0;
    image->format = XYBitmap;
    image->data = (char *) data;
    image->byte_order = LSBFirst;
    image->bitmap_unit = 8;
    image->bitmap_bit_order = LSBFirst;
    image->bitmap_pad = 8;
    image->depth = 1;
    image->bytes_per_line = (image->width + 7) / 8;

    return BitmapSuccess;
}


/*
 *  Dump weights in the specified file.
 */
DumpWeights(fname)
     char fname[];
{
  FILE  *outfile, *fopen();
  int  i, j;
  char realfname[80];

  /* Dump weights */

  sprintf (realfname, "%s.wts", fname);
  outfile = fopen ( realfname, "w" );
  fprintf (outfile, "%s\n", archfname);
  for (i=0; i<N; i++)
    for (j=0; j<N; j++)
      if ( Weights[i][j] != 0.0 )
	fprintf (outfile, "%d %d %f ", i, j, Weights[i][j]);

  fprintf (outfile, "%d %d %f ", -1, -1, -1.0);   /* Signal EOF */

  fclose (outfile);
}


/*
 *  Get weights from the specified file.  NOTE: re-reads archfname.
 */
GetWeights(fname)
     char fname[];
{
  FILE  *infile, *fopen();
  int  i, j;
  float inweight;
  char realfname[80];

  /* Get weights */
  
  sprintf (realfname, "%s.wts", fname);
  infile = fopen ( realfname, "r" );
  fscanf(infile, "%s", archfname);
  GetArchitecture(archfname);
  for (i=0; i<N; i++)
    for (j=0; j<N; j++)
      Weights[i][j] = 0.0;         /* Default weight */

  i = 11;                          /* Arbitrary +ive */
  while ( i >= 0 )
    {
      fscanf (infile, "%d %d %f", &i, &j, &inweight);
      if ( i >= 0 )
	Weights[i][j] = inweight;
    }

  fclose (infile);
}


InitializeGlobals()
{
  Epoch = 0;
  WeightRange = 0.5;
  SigmoidPrimeOffset = 0.1;
  HyperErr = 0;    /* was 1 */
  SplitEpsilon = 1;
  Epsilon = 1.0;
  Momentum = 0.0;
  ModeSwitchThreshold = 0.0;
  MaxFactor = 1.75;
  Decay = -0.0001;
  SinglePass = 0;
  SingleEpoch = 0;
  Step = 0;
  Restart = 1;
  KeepScore = 0;
  TotalError = 0.0;
  ScoreThreshold = 0.4;
  TotalErrorBits = 0;
  PosErrorBits = 0;
}



BuildDataStructures (ninputs, nhidden, noutputs)
     int ninputs, nhidden, noutputs;
{
  int i;

  Nunits      = 1 + ninputs + nhidden + noutputs;
  Ninputs     = ninputs;
  FirstHidden = 1 + ninputs;
  Nhidden     = nhidden;
  FirstOutput = 1 + ninputs + nhidden;
  Noutputs    = noutputs;

  for (i=0; i<=Nunits; i++)    Outputs[i] = 0.0;
  for (i=0; i<=Nunits; i++)    ErrorSums[i] = 0.0;
  for (i=0; i<=Nunits; i++)    Errors[i] = 0.0;
  for (i=0; i<=Nunits; i++)    Nconnections[i] = 0;
  
  Outputs[0] = 1.0;        /* The bias unit */
}


/*
 * Return a float between -range and +range.
 */
float RandomWeight (range)
     float range;
{
  return ( (float) (range * (random()%1000 / 500.0)) - range );
}


/*
 *  Build a connection from every unit in range1 to every unit in range2.
 *  Also add a connection from the bias unit (unit 0) to every unit in 
 *  range2.  Set up random weights on links.
 *
 *  NOTE:  If Nconnections[i] != 0 this means that the node already has
 *     some incoming links, and must be treated accordingly.
 */
ConnectLayers (start1, end1, start2, end2)
     int start1, end1, start2, end2;
{
  float RandomWeight()/*, fabs()*/;
  float rw;
  int i, j, k;

  Epoch = 0;
  
  for (i=start2; i<=end2; i++)
    {
      if (Nconnections[i] == 0)   /* First time, so do bias unit. */
	{
	  Nconnections[i] = 1;
	  Connections[i][0] = 0;
	  Weights[i][0] = RandomWeight(WeightRange);
	  DeltaWeights[i][0] = 0.0;
	  Slopes[i][0] = 0.0;
	  PrevSlopes[i][0] = 0.0;
	}
      
      for (j=start1; j<=end1; j++)
	{
	  k = Nconnections[i];
	  Connections[i][k] = j;

	  DeltaWeights[i][k] = 0.0;
	  Slopes[i][k] = 0.0;
	  PrevSlopes[i][k] = 0.0;

	  Nconnections[i]++;
	}
    }

  InitWeights();
}


/* 
 *  For each connection, select a random initial weight between WeightRange
 *  and its negative.  Clear delta and previous delta values.
 */
InitWeights()
{
  int i, j;

  for (i=0; i<Nunits; i++)
    for (j=0; j<Nconnections[i]; j++)
      {
	Weights[i][j] = RandomWeight(WeightRange);
	DeltaWeights[i][j] = 0.0;
	Slopes[i][j] = 0.0;
	PrevSlopes[i][j] = 0.0;
      }
}


/*
 *  Save the current slope values as PrevSlopes, and "clear" all current
 *  slopes (actually set to corresponding weight, decayed a bit).
 */
ClearSlopes()
{
  int i, j;

  for (i=FirstHidden; i<Nunits; i++)
    for (j=0; j<Nconnections[i]; j++)
      {
	PrevSlopes[i][j] = Slopes[i][j];
	Slopes[i][j] = ( Decay * Weights[i][j] );
      }
}


float Sigmoid (activation)
     float activation;
{
  if (activation < -15.0) 
    return (0.0);  
  else if (activation >  15.0) 
    return (1.0);
  else
    return ( 1.0 / (1.0 + exp(-activation)) );
}


float SigmoidPrime (output)
     float output;
{
  return ( SigmoidPrimeOffset + (output * (1.0 - output)) );
}



/*
 *  Node function, varies from node to node.
 */
float NodeFunction(nix)
     int nix;               /* Node index  */
{
  float sum, scan_sum, sig_sum;
  float SigmoidPrime();
  int i,j,k, ix;

  /* For angle RBF node, dist_squared = (sin(c)-sin)^2 + (cos(c)-cos)^2,
   *   return exp(-dist_squared/sig_squared), a gaussian.
   *
   *   Weights[nix][1] is sin(c),  Outputs[Connections[nix][3]] is sin,
   *   Weights[nix][2] is cos(c),  Outputs[Connections[nix][4]] is cos.
   *   Weights[nix][5] is sigma(c).
   */

  if (AngleNode[nix])
    {
      DistSquared[nix] = 
	(((Weights[nix][1]-Outputs[Connections[nix][3]])*
	  (Weights[nix][1]-Outputs[Connections[nix][3]]))  +
	 ((Weights[nix][2]-Outputs[Connections[nix][4]])*
	  (Weights[nix][2]-Outputs[Connections[nix][4]])));
      return (exp (-DistSquared[nix]/(Weights[nix][5]*Weights[nix][5])));
    }

  /* Averaging nodes take an unweighted average of their inputs
   *  (weights are constrained to be zero).   
   */

  else if (AveragingNode[nix])
    {
      sum = 0.0;
      for (i=1; i<Nconnections[nix]; i++)
	sum += ( Outputs[ Connections[nix][i] ] * Weights[nix][i] );
      return (sum / (float) Nconnections[nix]);
    }


  /*   Scanning nodes:  these guys center themselves over each point in
   *    the LM interior in turn, and are interested in a point (i.e.
   *    gated on) iff the TR boundary is on at that point.  They return
   *    the average response over all "interesting" positions.
   *
   *   MaxScanningNodes are similar, but return the max response.
   */

  else if ((ScanningNode[nix]) || (MaxScanningNode[nix]))
    {

      for (ix=1; ix<=5; ix++)
	AvgOutputs[nix][ix] = 0.0;   

      for (ix=1; ix<=5; ix++)
	ScanningSum[nix][ix] = 0.0;  

      MaxResponse[nix] = 0.0;
      for (ix=1; ix<=5; ix++)
	OutputsforMaxResponse[nix][ix] = 0.0;   

      scan_sum = 0.0;
      for (ix=0; ix<NumTRBoundaryPoints[SceneIndex]; ix++)
	{
	  i = TRBoundaryPoints[SceneIndex][ix][0];
	  j = TRBoundaryPoints[SceneIndex][ix][1];

	  /* Look at position, load bottom (input) units. */
	      
	  Outputs[1] = PeekInterior(SceneIndex, i-1,j);
	  Outputs[2] = PeekInterior(SceneIndex, i,j-1);
	  Outputs[3] = PeekInterior(SceneIndex, i,j);
	  Outputs[4] = PeekInterior(SceneIndex, i+1,j);
	  Outputs[5] = PeekInterior(SceneIndex, i,j+1);

	  for (k=1; k<=5; k++)
	    AvgOutputs[nix][k] = AvgOutputs[nix][k]+Outputs[k];

	  sum = 0.0;
	  for (k=0; k<Nconnections[nix]; k++)
	    sum += ( Outputs[ Connections[nix][k] ] * Weights[nix][k] );
	  sig_sum = Sigmoid(sum);
	  scan_sum = scan_sum + sig_sum;
	  if (sig_sum > MaxResponse[nix])
	    {
	      MaxResponse[nix] = sig_sum;
	      for (ix=1; ix<=5; ix++)
		OutputsforMaxResponse[nix][ix] = Outputs[ix];
	    }

	  for (k=1; k<=5; k++)
	    ScanningSum[nix][k] = 
	      (ScanningSum[nix][k] + 
	       (SigmoidPrime(sig_sum) * Outputs[k]));
	}

      for (ix=1; ix<=5; ix++)
	AvgOutputs[nix][ix] = 
	  (AvgOutputs[nix][ix] / (float) NumTRBoundaryPoints[SceneIndex]);

      if (ScanningNode[nix])
	return (scan_sum / (float) NumTRBoundaryPoints[SceneIndex]);
      else      /* MaxScanningNode */
	return (MaxResponse[nix]);
    }


  /* Normal sigmoid of weighted sum */

  else 
    {
      sum = 0.0;
      for (i=0; i<Nconnections[nix]; i++)
	sum += ( Outputs[ Connections[nix][i] ] * Weights[nix][i] );
      return (Sigmoid(sum));
    }
}  


/*
 *  Compute the error for one output unit.  
 *  HyperErr==0 => use squared error.
 *  HyperErr==1 => use atanh.
 */
float ErrorFunction (desired, actual, curpos,this)
     float desired, actual;
     int curpos;  /* Current positive instance (0 - Noutputs-1) */
     int this;    /* Index of this output node */
{
  float dif, uncut_dif;

  /*
   *  UPDATED!
   *
   *    If the desired output is -ive, this indicates don't-care,
   *    so the error is 0.
   *
   *    Notice that if this is a negative instance, we divide the 
   *    perceived difference between desired and actual by ErrorCut
   *    (default ErrorCut = 1.0).   UNLESS this is an antonym of the
   *    current spatial concept, and we care about that (DoAntonyms).
   */

  dif = ((desired < 0.0) ? 0.0 : desired - actual);

  if ( KeepScore )   
    {
      TotalError += dif*dif;
      if ( fabs(dif) >= ScoreThreshold )
	{
	  TotalErrorBits++;
	  if (desired > 0.5) 
	    PosErrorBits++; 
	}
    }

  if ( HyperErr == 0 )         /* Not using atanh for error */
    {
      if ((-0.00001 < dif) && (dif < 0.00001))    /* was 0.1 */
	return (0.0);
      else 
	return (dif*Beta[curpos][this]);
    }
  else                         /* Using atanh for error */
    {
      if ( dif < -.9999999 )
	return (-17.0);
      else if ( dif > .9999999 )
	return (17.0);
      else
	return ( log ( (1.0+dif) / (1.0-dif) ) * Beta[curpos][this] );
    }
}


/*
 *  This is it, ya Habaayib:  the forward pass in BP.
 */
ForwardPass (input)
     float input[];
{
  register int i, j;
  float sum, NodeFunction();
  
  /*  
   *  Kill the bias term for angle nodes, and freeze weights coming
   *    from sin,cos.
   *
   *  Also, if incoming canonical sin,cos weights are out of bounds,
   *    correct for this.
   */

  for (i=FirstHidden; i<FirstHidden+Nhidden; i++)
    if (AngleNode[i])
      {
	Weights[i][0] = 0.0;     /* Bias */

	Weights[i][3] = 1.0;     /* Incoming sin,cos */
	Weights[i][4] = 1.0;
      }

  /*  Freeze incoming weights to averaging nodes. */

  else if (AveragingNode[i])
    {
      for (j=1; j<Nconnections[i]; j++)
	Weights[i][j] = 1.0;
    }


  /*
   *  Tie weights.  Assuming Weights[Tied[i][0][0]][Tied[i][0][1]] has
   *   already been set (first pair of class).
   */

  for (i=0; i<NumTied; i++)
    for (j=0; j<TiedClassSize[i]; j++)
      Weights[Tied[i][j][0]][Tied[i][j][1]] =
	Weights[Tied[i][0][0]][Tied[i][0][1]];


  /* Load in the input vector */

  for (i=0; i<Ninputs; i++)
    Outputs[i+1] = input[i];


  /* 
   *  Some weights take the value of certain input nodes.  
   *   Set these here. 
   */

  for (i=0; i<NumInputWeights; i++)
    Weights[InputWeights[i][0]][InputWeights[i][1]] = 
      Outputs[InputWeights[i][2]];


  /* For each unit, determine its output */

  for (j=FirstHidden; j<Nunits; j++)
    Outputs[j] = NodeFunction(j);
}


/*
 *  Goal is a vector of desired values for the output units.  Propagate
 *  the error back through the net, accumulating weight deltas.
 */
BackwardPass (goal)
     float goal[];
{
  register int i, j, cix;     /* cix is "connection index" */
  int hix, a_flag;   /* hidden unit index, antonym flag */
  float deriv;       /* deriv of netinput wrt weight */
  int curpos;        /* current positive */
  float AngleSlope();


  /* Compute error sums for output and other nodes */

  if (ExplicitNegatives == TRUE)
    {
      for (i=FirstOutput; i<Nunits; i++)   
	ErrorSums[i] = ErrorFunction( goal[i-FirstOutput], Outputs[i], 
				     i-FirstOutput, i-FirstOutput);
    }

  else            /* Implicit negs, so must be a +ive somewhere */
    {
      curpos = -1;

      for (i=FirstOutput; (i<Nunits && curpos == -1); i++)   
	if (goal[i-FirstOutput] > 0.5) curpos = i-FirstOutput;

      for (i=FirstOutput; i<Nunits; i++)   
	ErrorSums[i] = ErrorFunction( goal[i-FirstOutput], Outputs[i], 
				     curpos, i-FirstOutput);
    }

  for (i=0; i<FirstOutput; i++)
    ErrorSums[i] = 0.0;

  /* Back-prop.  When we reach a given unit in loop, error from all later
   * units will have been collected.
   */
  for (j=Nunits-1; j>=FirstHidden; j--)
    {

      /* Start by loading output units for RF with avg value over 
       *    TR-gated positions.  Use this when computing Slopes.
       *    If MaxScanningNode, fill w/ those outputs which gave the
       *    greatest overall response.
       */   
	      
      if (ScanningNode[j])
	for (i=1; i<=5; i++)
	  Outputs[i] = AvgOutputs[j][i];

      if (MaxScanningNode[j])
	for (i=1; i<=5; i++)
	  Outputs[i] = OutputsforMaxResponse[j][i];


      /* Errors[j] = dE/dnet(j), inapplicable for angle nodes, since they
       *     don't take net input.
       * ErrorSums[j] = -dE/do(j).  This gets calculated for angle
       *     nodes by the guys above them, and then gets used in AngleSlope.
       *     Note that for nodes right below angle nodes, ErrorSums is
       *     calculated using AngleSlopes, instead of the usual method.
       */

      /*  AveragingNodes have a different derivative than sigmoids! */
      /*  Since SplitEpsilon=1 takes care of division, don't divide by
	  Nconnections[j] here.	  */

      if (AveragingNode[j] == TRUE)
	Errors[j] = /* (1.0 / (float) Nconnections[j]) *   */ ErrorSums[j];
      else
	Errors[j] = SigmoidPrime(Outputs[j]) * ErrorSums[j];

      for (i=0; i<Nconnections[j]; i++)
	{
	  cix = Connections[j][i];
	  if (AngleNode[j])
	    ErrorSums[cix] += AngleSlope(j,i);
	  else
	    ErrorSums[cix] += ( Errors[j] * Weights[j][i] );

	  if (AngleNode[j] == TRUE)
	    Slopes[j][i] += AngleSlope(j,i);
	  else if (ScanningNode[j])      /* Special case, see thesis.  */
	    Slopes[j][i] += 
	      ((ErrorSums[j]*(1.0/(float)NumTRBoundaryPoints[SceneIndex])) *
	       (ScanningSum[j][i]));
	  else
	    Slopes[j][i] += ( Errors[j] * Outputs[cix] );
	}
    }


  /* Weight tying */

  for (i=0; i<NumTied; i++)
    for (j=1; j<TiedClassSize[i]; j++)
      Slopes[Tied[i][0][0]][Tied[i][0][1]] +=
	Slopes[Tied[i][j][0]][Tied[i][j][1]];

  for (i=0; i<NumTied; i++)
    Slopes[Tied[i][0][0]][Tied[i][0][1]] =
      (Slopes[Tied[i][0][0]][Tied[i][0][1]] / (float)TiedClassSize[i]);
}


/*
 *  Returns slope for a link coming into an angle (RBF) node.
 *
 *  Deriv is df/dx, whatever x winds up being.  Most of the job is 
 *   to compute -dE/df * df/dx = -dE/dx, i.e. minus the slope.
 *   (ErrorSums = -dE/df).   
 *
 *  Note that deriv gets / by NTrainingPatterns, since this gets summed
 *    over all patterns (i.e. we wind up taking the average).
 *
 *  Updated so now constrains canonical sin and cos to be legal 
 *   (sin^2 + cos^2 = 1).  This is reflected in derivg, which is 
 *   dg/dx, whatever x is.  Then we return (-dE/dx + -dg/dx).
 *
 *  Another update:  sigma held high 
 */
float AngleSlope(node,in_line)
     int node, in_line;
{
  float deriv, derivg, sig_squared;

  sig_squared = Weights[node][5]*Weights[node][5];
  derivg = 0.0;               /* Default */

  if (in_line == 1)           /* weight = sin (canonical angle), df/dsinc */
    {
      sig_squared = SigmaDeriv*SigmaDeriv;  /* use imposed value instead */
      deriv = 
	(-2 * (Weights[node][1]-Outputs[Connections[node][3]])) /
	  (exp(DistSquared[node]/sig_squared) * sig_squared);
      /* dg/dsinc = 4 sinc (sinc^2 + cosc^2 - 1) */
/*      derivg = 
	(4 * 
	 Weights[node][1] *
	 ((Weights[node][1]*Weights[node][1]) +
	  (Weights[node][2]*Weights[node][2]) - 1.0));
*/
    }

  else if (in_line == 2)      /* weight = cos (canonical angle), df/dcosc */
    {
      sig_squared = SigmaDeriv*SigmaDeriv;  /* use imposed value instead */
      deriv = 
	(-2 * (Weights[node][2]-Outputs[Connections[node][4]])) /
	  (exp(DistSquared[node]/sig_squared) * sig_squared);
      /* dg/dcosc = 4 cosc (sinc^2 + cosc^2 - 1) */
/*
      derivg = 
	(4 * 
	 Weights[node][2] *
	 ((Weights[node][1]*Weights[node][1]) +
	  (Weights[node][2]*Weights[node][2]) - 1.0));
*/
    }

  else if (in_line == 3)      /* weight = 1.0 from sin. for df/dsin */
    deriv = 
      (2 * (Weights[node][1]-Outputs[Connections[node][3]])) /
	(exp(DistSquared[node]/sig_squared) * sig_squared);

  else if (in_line == 4)      /* weight = 1.0 from cos. for df/dcos */
    deriv = 
      (2 * (Weights[node][2]-Outputs[Connections[node][4]])) /
	(exp(DistSquared[node]/sig_squared) * sig_squared);

  else if (in_line == 5)      /* weight = sigma for RBF */
    deriv = 
      (2 * DistSquared[node]) /
	(exp(DistSquared[node]/sig_squared) * sig_squared*Weights[node][5]);
      
  else
    deriv = 0.0;

  deriv = deriv / NTrainingPatterns;

  return ((ErrorSums[node] * deriv) + (Atten*-derivg));
}


/*
 *  Update all weights in the network as a function of each weight's current
 *  slope, previous slope, and the size of the last jump.
 */
UpdateWeights()
{
  int i, j;
  float next_step, shrink_factor;

  shrink_factor = MaxFactor / ( 1.0 + MaxFactor );

  for (j=FirstHidden; j<Nunits; j++)
    for (i=0; i<Nconnections[j]; i++)
      {
	next_step = 0.0;

	if ( AngleNode[j] )
	  {
	    /* For now, update sinc, cosc, and sigma links. */
	    if ((i == 1) || (i == 2) || (i == 5))
	      next_step += ((SplitEpsilon ? 
			     ( (Epsilon * Slopes[j][i]) / Nconnections[j] ) :
			     ( Epsilon * Slopes[j][i] ))
			    + (Momentum * DeltaWeights[j][i]) );

	  }

	else if ( DeltaWeights[j][i] > ModeSwitchThreshold  )
	  {                                 /* Last step was signif. +ive.. */
	    if ( Slopes[j][i] > 0.0 )       /* Add in epsilon if +ive slope */
	      next_step += (SplitEpsilon ? 
			    ( (Epsilon * Slopes[j][i]) / Nconnections[j] ) :
			    ( Epsilon * Slopes[j][i] ));
	    /* If slope > (or close to) prev slope, take max size step. */
	    if ( Slopes[j][i] > (shrink_factor * PrevSlopes[j][i]) )
	      next_step += ( MaxFactor * DeltaWeights[j][i] );
	    else        /*  Use quadratic estimate */
	      next_step += ( (Slopes[j][i]/(PrevSlopes[j][i]-Slopes[j][i])) 
			    * DeltaWeights[j][i] );
	  }
	else if ( DeltaWeights[j][i] < -ModeSwitchThreshold  )
	  {                                 /* Last step was signif. -ive. */
	    if ( Slopes[j][i] < 0.0 )       /* Add in epsilon if -ive slope */
	      next_step += (SplitEpsilon ? 
			    ( (Epsilon * Slopes[j][i]) / Nconnections[j] ) :
			    ( Epsilon * Slopes[j][i] ));
	    /* If slope < (or close to) prev slope, take max size step. */
	    if ( Slopes[j][i] < (shrink_factor * PrevSlopes[j][i]) )
	      next_step += ( MaxFactor * DeltaWeights[j][i] );
	    else        /*  Use quadratic estimate */
	      next_step += ( (Slopes[j][i]/(PrevSlopes[j][i]-Slopes[j][i])) 
			    * DeltaWeights[j][i] );
	  }
	else       /* Normal gradient descent, complete with momentum */
	  {
	    DidGradient++;
	    next_step += ((SplitEpsilon ? 
			   ( (Epsilon * Slopes[j][i]) / Nconnections[j] ) :
			   ( Epsilon * Slopes[j][i] ))
			  + (Momentum * DeltaWeights[j][i]) );
	  }
	
	/* Set delta weight, and adjust the weight itself. */

	if (Fixed[j][i] == FALSE)
	  {
	    DeltaWeights[j][i] = next_step;
	    Weights[j][i] += next_step;
	  }
      }
}


/*
 *  Perform forward and back propagation once for each pattern in the
 *  training set, collecting deltas.  Then burn in the weights.
 */
TrainOneEpoch()
{
  int pat, i,j,k;
  float x,y;
  float divisor, AspectRatio();

  ClearSlopes();

  for (SceneIndex=0; SceneIndex<NTrainingPatterns; SceneIndex++)
    {
      FillTinputs();
      ForwardPass( Tinputs );
      BackwardPass ( TrainingOutputs[SceneIndex] );
    }

  UpdateWeights();
  Epoch++;
}


/*
 *  Train the network for the specified number of epochs, printing out
 *  performance stats every 10 epochs.
 */
Train ( times )
     int times;
{
  int i, j, report, stopnow, anode;

  report = 10;

  stopnow = FALSE;
  for (i=0; ((i<times) && (stopnow == FALSE)); i++)
    {
      if ( Epoch % report == 0 )     /* Time to report status */
	{
	  anode = -1;
	  for (j=0; ((j<Nunits) && (anode==-1)); j++)
	    if (AngleNode[j] == TRUE)
	      anode = j;

	  DidGradient = 0;
	  KeepScore = 1;
	  TotalError = 0.0;
	  TotalErrorBits = 0;
	  PosErrorBits = 0;
	  TrainOneEpoch();
	  printf ("Epoch %d:  %d (%d) Bits Wrong, Total Error = %f, DidGradient = %d.\n",
		  (Epoch - 1), TotalErrorBits, PosErrorBits, 
		  TotalError, DidGradient);
	  if (((ZeroStop) && (TotalErrorBits == 0)) ||
	      ((PosZeroStop) && (PosErrorBits == 0)))
	    stopnow = TRUE;
	  KeepScore = 0;
	}
      else
	TrainOneEpoch();	
    }
}


Test ()
{
  int i,ix;
  
  SceneIndex = 4;;          /* Initial nonzero value */

  while (SceneIndex >= 0)
    {
      printf ("Enter the scene number [less than 0.0 => quit]: ");
      scanf("%d", &SceneIndex);
      
      if (SceneIndex >= 0)
	{
	  ForwardPass( Tinputs );

	  printf ("Output is: ");
	  for (i=0; i<Noutputs; i++)
	    printf ("%f  ", Outputs[FirstOutput+i]);
	  printf ("\n");
	}
    }
}


/*
 *  Fill the Tinputs vector preparatory to training.
 *  Uses the global variable SceneIndex.
 */
FillTinputs()
{
  float midheadx,midheady,midfootx,midfooty,dhx,dhy,vhx,vhy,dfx,dfy,vfx,vfy;
  /* float rint();*/
  int i,j, k, ii,jj;
  char c;
  Angle *directionptr;

  directionptr = (Angle *) malloc (sizeof(Angle));

  /*  
   *  Note that Tinputs[0] through Tinputs[4] are reserved 
   *     for scanning stuff.   Don't use them.   See NodeFunction 
   *     for details.
   */


  /*  Set up unit inputs for sin,cos, sigmas.    Input weights 
   *    should be numbers 1, 2, 5 for sin, cos, sigma respectively.
   */

  Tinputs[5] = 1.0;         /*  weight = sin(theta)  */
  Tinputs[6] = 1.0;         /*  weight = cos(theta)  */
  Tinputs[7] = 1.0;         /*  weight = sigma  */

  /* Load in sin,cos for proximal orientation. */

  Tinputs[8] = ProxOrientation[SceneIndex].Sine;
  Tinputs[9] = ProxOrientation[SceneIndex].Cosine;

  /*  Major axis orientation sin,cos.  Using only [0] for now
   *     ([1] holds the other possibility in ambiguous cases;
   *      will probably want to change this representation).
   */

  Tinputs[10] = MajorAxisOrientation[SceneIndex][0].Sine;
  Tinputs[11] = MajorAxisOrientation[SceneIndex][0].Cosine;
}
